/* Wejp's XMMS Serial Remote Control Plugin
 *
 * Copyright (C) 2003 Johannes Heimansberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#define VERSION_NUMBER "0.1.1"
#define PLUGIN_NAME "Wejp's Serial Remote Control " VERSION_NUMBER

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <pthread.h>

#include <xmms/plugin.h>
#include <xmms/xmmsctrl.h>
#include <xmms/configfile.h>
#include <xmms/util.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>

#include <termio.h>

static pthread_t rm_thread;
static int thread_running;

static int vol_l, vol_r;
static int mute_flag = 0;

enum { PLAY_PAUSE,
       STOP,
       NEXT,
       PREV,
       VOL_UP,
       VOL_DOWN,
       QUIT,
       MUTE,
       REPEAT,
       SHUFFLE,
       DELETE };

static char function_name[11][32] = { "Play/Pause", 
                                      "Stop",
                                      "Next",
                                      "Previous",
                                      "Volume Up",
                                      "Volume Down",
                                      "Quit",
                                      "Mute",
                                      "Repeat",
                                      "Shuffle",
                                      "Delete" };

static char serial_device[128] = "/dev/ttyS0";
static int  key[5];

static void init();
static void cleanup();
static void about();

GeneralPlugin rm_gp =
{
	NULL,
	NULL,
	-1,
	PLUGIN_NAME,
	init,
	about,
	NULL,
	cleanup
};

/* Checks for key presses on the remote control and returns the number of the pressed key */
static int get_remote_key()
{
	int remote_key = 0;
	int fd = 0;
	unsigned int status = 0;

	fd = open(serial_device, O_RDWR | O_NOCTTY); /* Open serial port */
	if (fd > 0)
	{
		/* Read status bits */
		ioctl(fd, TIOCMGET , &status);
		close(fd);
		if 	    (status & TIOCM_CD  && !(status & TIOCM_DSR || status & TIOCM_CTS || status & TIOCM_RI))  remote_key = 1;
		else if	(status & TIOCM_DSR && !(status & TIOCM_CD  || status & TIOCM_CTS || status & TIOCM_RI))  remote_key = 2;
		else if	(status & TIOCM_CTS && !(status & TIOCM_CD  || status & TIOCM_DSR || status & TIOCM_RI))  remote_key = 3;
		else if	(status & TIOCM_RI  && !(status & TIOCM_CD  || status & TIOCM_DSR || status & TIOCM_CTS)) remote_key = 4;
	}
	else
		remote_key = -1; /* Error. Serial port could not be opened. */
	return remote_key;
}

static int load_config()
{
	FILE *file;
	char *path_to_config_file, *home_directory, ch;
	char buffer[128] = "";
	int  bufcnt = 0, keynum = 0;

	home_directory = getenv("HOME");
	path_to_config_file = (char*) malloc(strlen(home_directory) + 128);
	sprintf(path_to_config_file, "%s/.xmms/wsremote.conf", home_directory);
	/* Read config file... */
	file = fopen(path_to_config_file, "r");
	if (file)
	{
		while(!feof(file))
		{
			ch = fgetc(file);
			/* Skip comments (#)... */
			if (ch == '#')
				while (ch != '\n' && ch != '\r') ch = fgetc(file);
			/* Blanks too... */
			if (ch == ' ')
				while (ch == ' ') ch = fgetc(file);
			bufcnt = 0;
			while (ch != ' ' && ch != '\n' && ch != '\r' && !feof(file) && bufcnt < 126)
			{
				buffer[bufcnt] = ch;
				bufcnt++;
				ch = fgetc(file);
			}
			buffer[bufcnt] = '\0';
			bufcnt = 0;
			if (keynum == -42) /* Serielles Geraet festlegen... */
			{
				strncpy(serial_device, buffer, 126);
				keynum = 0;
			}
			if (keynum > 0)
			{
				if (strcmp(buffer, "PlayPause") == 0)  key[keynum] = PLAY_PAUSE;
				if (strcmp(buffer, "Next") == 0)       key[keynum] = NEXT;
				if (strcmp(buffer, "Prev") == 0)       key[keynum] = PREV;
				if (strcmp(buffer, "VolumeUp") == 0)   key[keynum] = VOL_UP;
				if (strcmp(buffer, "VolumeDown") == 0) key[keynum] = VOL_DOWN;
				if (strcmp(buffer, "Shuffle") == 0)    key[keynum] = SHUFFLE;
				if (strcmp(buffer, "Repeat") == 0)     key[keynum] = REPEAT;
				if (strcmp(buffer, "Quit") == 0)       key[keynum] = QUIT;
				if (strcmp(buffer, "Delete") == 0)     key[keynum] = DELETE;
				keynum = 0;
			}
			if (strcmp(buffer, "Key1") == 0) keynum = 1;
			if (strcmp(buffer, "Key2") == 0) keynum = 2;
			if (strcmp(buffer, "Key3") == 0) keynum = 3;
			if (strcmp(buffer, "Key4") == 0) keynum = 4;
			if (strcmp(buffer, "RemoteDevice") == 0) keynum = -42;
		}
		fclose(file);
	}
	else
	{
		printf("WSRC-Plugin: No config file found!\n");
		return -2;
	}
	free(path_to_config_file);
	return 0;
}

static void handle_remote_input(int keynum)
{
	if (key[keynum] == PREV)
	{
		xmms_remote_playlist_prev(rm_gp.xmms_session);
	}
	if (key[keynum] == NEXT)
	{
		xmms_remote_playlist_next(rm_gp.xmms_session);
	}
	if (key[keynum] == PLAY_PAUSE)
	{
		if (xmms_remote_is_playing(rm_gp.xmms_session))
			xmms_remote_pause(rm_gp.xmms_session);
		else
			xmms_remote_play(rm_gp.xmms_session);
	}
	if (key[keynum] == STOP)
	{
		xmms_remote_stop(rm_gp.xmms_session);
	}
	if (key[keynum] == MUTE)
	{
		if (mute_flag)
		{
			mute_flag = 0;
			xmms_remote_set_volume(rm_gp.xmms_session, vol_l, vol_r);
		} 
		else
		{
			mute_flag++;
			xmms_remote_get_volume(rm_gp.xmms_session, &vol_l, &vol_r);
			xmms_remote_set_volume(rm_gp.xmms_session, 0, 0);
		}
	}
	if (key[keynum] == VOL_UP)
	{
		if (mute_flag)
		{
			mute_flag = 0;
			xmms_remote_set_volume(rm_gp.xmms_session, vol_l, vol_r);
		}
		else
		{
			xmms_remote_get_volume(rm_gp.xmms_session, &vol_l, &vol_r);
		}
		vol_l += 5;
		vol_r += 5;
		if (vol_l > 100) vol_l = 100;
		if (vol_r > 100) vol_r = 100;
		xmms_remote_set_volume(rm_gp.xmms_session, vol_l, vol_r);
    }
    if (key[keynum] == VOL_DOWN)
    {
		if (mute_flag)
		{
			mute_flag = 0;
			xmms_remote_set_volume(rm_gp.xmms_session, vol_l, vol_r);
		}
		else
		{
			xmms_remote_get_volume(rm_gp.xmms_session, &vol_l, &vol_r);
		}
		vol_l -= 5;
		vol_r -= 5;
		if (vol_l < 0) vol_l = 0;
		if (vol_r < 0) vol_r = 0;
		xmms_remote_set_volume(rm_gp.xmms_session, vol_l, vol_r);
	}
	if (key[keynum] == QUIT)
	{
		xmms_remote_quit(rm_gp.xmms_session);
	}
	if (key[keynum] == REPEAT)
	{
		xmms_remote_toggle_repeat(rm_gp.xmms_session);
	}
	if (key[keynum] == SHUFFLE)
	{
		xmms_remote_toggle_shuffle(rm_gp.xmms_session);
	}
	if (key[keynum] == DELETE)
	{
		int playlist_length = xmms_remote_get_playlist_length(rm_gp.xmms_session);
		if (playlist_length > 0)
		{
			int i;
			i = xmms_remote_get_playlist_pos(rm_gp.xmms_session);
			xmms_remote_playlist_delete(rm_gp.xmms_session, i);
		}
	} 
}

static void *mainloop()
{
	int keynum = 0;

	thread_running = 1;
	while(thread_running == 1)
	{
		keynum = get_remote_key();
		if (keynum > 0)
		{
			handle_remote_input(keynum);
			/* wait for 0.25 a seconds... */
			usleep(250000);
		}
		else if (keynum < 0) /* Error */
		{
			printf("WSRC-Plugin Error: Cannot open serial port!\n");
			break;
		}
		usleep(15);
	}
	return NULL;
}

static void init()
{
	if (load_config() == 0) /* If config file was succesfully loaded start the plug-in. */
	{
		pthread_create(&rm_thread, NULL, mainloop, NULL);
	}
}

static void cleanup()
{
	thread_running = 0;
	pthread_join(rm_thread, NULL);
}

GeneralPlugin *get_gplugin_info(void)
{
	return &rm_gp;
}

static void about()
{
	static GtkWidget *dialog;
	char   *message, msg[1024], key_mapping[1024] = "";

	if (dialog != NULL)
		return;

	sprintf(key_mapping, "Key1=%s\nKey2=%s\nKey3=%s\nKey4=%s\n",
	        function_name[key[1]], function_name[key[2]], function_name[key[3]], function_name[key[4]]);

	sprintf(msg,
	        "Wejp's Serial Remote Control Plugin %s\n\n "
	        "Copyright (c) 2003 Johannes Heimansberg\n"
	        "http://www.wejp.de/xmmsremote/\n\n"
	        "Selected serial port: %s\n\n"
	        "Current key configuration:\n%s\n "
	        "Config file: ~/.xmms/wsremote.conf\n\n"
	        "This program is free software, released under the GNU General Public "
	        "License.",
	        VERSION_NUMBER,
	        serial_device,
	        key_mapping);

	message = msg;

	dialog = xmms_show_message("About Wejp's Serial Remote Control Plugin",	message, "Ok", FALSE, NULL, NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);
}
